import pandas as pd
import json
import os

METRICS_LIST = [
    "files",
    "functions",
    "complexity",
    "comment_lines_density",
    "duplicated_lines_density",
    "coverage",
    "ncloc",
    "tests",
    "test_errors",
    "test_failures",
    "test_execution_time",
    "security_rating",
]


def read_json(json_path):
    with open(json_path) as json_file:
        json_obj = json.load(json_file)

    return json_obj


def create_base_component_df(json_list):
    df = pd.DataFrame()

    for i in json_list:
        base_component = read_json(i)

        base_component_data = base_component["baseComponent"]["measures"]

        base_component_df = pd.DataFrame(base_component_data)

        base_component_df["filename"] = os.path.basename(i)

        df = pd.concat([df, base_component_df], ignore_index=True)

    return df


def metric_per_file(json):
    file_json = []

    for component in json["components"]:
        file_json.append(component)

    return file_json


def check_component_is_valid(component, language_extension):
    return (
        component["qualifier"] == "DIR" or component["language"] == language_extension
    )


def generate_file_dataframe_per_release(metric_list, json, language_extension):
    df_columns = metric_list + ["qualifier"]
    df = pd.DataFrame(columns=df_columns)

    for file in json:
        if check_component_is_valid(file, language_extension):
            df.at[file["path"], "qualifier"] = file["qualifier"]
            for measure in file["measures"]:
                df.at[file["path"], measure["metric"]] = measure["value"]

    df.reset_index(inplace=True)
    df = df.rename({"index": "path"}, axis=1).drop(["files"], axis=1)

    return df


def create_file_df(json_list, language_extension="js"):
    df = pd.DataFrame()

    for json_file in json_list:
        file_component = read_json(json_file)

        file_component_data = metric_per_file(file_component)

        file_component_df = generate_file_dataframe_per_release(
            METRICS_LIST, file_component_data, language_extension=language_extension
        )

        file_component_df["filename"] = os.path.basename(json_file)

        df = pd.concat([df, file_component_df], ignore_index=True)

    return df



class MeasureSoftGramCoreException(Exception):
    """Base MeasureSoftGram Core exception"""

    pass


class InvalidMetricValue(MeasureSoftGramCoreException):
    """Raised when a invalid metric value is provided to the interpretation function"""

    pass


class InvalidInterpretationFunctionArguments(MeasureSoftGramCoreException):
    """Raised when invalid arguments are provided to an interpretation function"""

    pass


class ValuesAndWeightsOfDifferentSizes(MeasureSoftGramCoreException):
    """Raised when the length of the weight and values are not equal"""

    pass


class ImplicitMetricValueError(MeasureSoftGramCoreException):
    """
    Exception levantanda quando uma métrica não é passada e não é possível
    obter seu valor de maneira implícita.
    """

    pass


class InvalidThresholdValue(MeasureSoftGramCoreException):
    """Raised when a invalid threshold value is provided to the interpretation function"""

    pass


class InvalidGainInterpretationValue(MeasureSoftGramCoreException):
    """Raised when a invalid gain interpretation value is provided to the interpretation function"""

    pass


class InvalidCheckThreshold(MeasureSoftGramCoreException):
    """Raised when a invalid check threshold function is called"""

    pass


class MeasureKeyNotSupported(MeasureSoftGramCoreException):
    """Raised when a measure key is not supported"""

    pass


class ReleasePlannedAndDevelopedOfDifferentSizes(MeasureSoftGramCoreException):
    """Raised when the sizes of the planned and developed release vectors are different"""

    pass



import numpy as np
from marshmallow.exceptions import ValidationError

from core.schemas import (
    CalculateCharacteristicSchema,
    CalculateMeasureSchema,
    CalculateSubCharacteristicSchema,
    CalculateTSQMISchema,
)
from core.transformations import calculate_aggregated_weighted_value
from util.constants import AGGREGATED_NORMALIZED_MEASURES_MAPPING
from util.exceptions import MeasureKeyNotSupported


def calculate_measures(
    extracted_measures: CalculateMeasureSchema,
    config: dict = {"characteristics": [{"subcharacteristics": [{"measures": [{"key": ""}]}]}]},
):
    # Validate if outter keys is valid
    try:
        data = CalculateMeasureSchema().load(extracted_measures)
    except ValidationError as error:
        raise ValidationError(f"error: Failed to validate input.\nschema_errors: {error.messages}")

    # Objeto retornado em caso de sucesso
    result_data = {"measures": []}

    valid_measures = AGGREGATED_NORMALIZED_MEASURES_MAPPING.keys()

    for measure in data["measures"]:
        measure_key: str = measure["key"]

        if measure_key not in valid_measures:
            raise MeasureKeyNotSupported(f"Measure {measure_key} is not supported")

        measure_params = measure["parameters"]
        schema = AGGREGATED_NORMALIZED_MEASURES_MAPPING[measure_key]["schema"]

        try:
            validated_params = schema().load(measure_params)
        except ValidationError as exc:
            raise ValidationError(
                f"error: Metric parameters {measure_key} are not valid.\nschema_errors: {exc.messages}"
            )

        aggregated_normalized_measure = AGGREGATED_NORMALIZED_MEASURES_MAPPING[measure_key][
            "aggregated_normalized_measure"
        ]

        measures = [
            measure
            for characteristic in config["characteristics"]
            for subcharacteristic in characteristic["subcharacteristics"]
            for measure in subcharacteristic["measures"]
        ]

        threshold_config = {
            key: value
            for measure in measures
            for key, value in measure.items()
            if measure["key"] == measure_key and ("min_threshold" == key or "max_threshold" == key)
        }
        result = aggregated_normalized_measure(validated_params, **threshold_config)

        result_data["measures"].append(
            {
                "key": measure_key,
                "value": result,
            }
        )

    return result_data


def calculate_subcharacteristics(extracted_subcharacteristics):
    try:
        data = CalculateSubCharacteristicSchema().load(extracted_subcharacteristics)
    except ValidationError as error:
        raise ValidationError(f"error: Failed to validate input.\nschema_errors: {error.messages}")

    result_data = {"subcharacteristics": []}

    for subcharacteristic in data["subcharacteristics"]:
        subcharacteristic_key: str = subcharacteristic["key"]

        vector_aggregated_normalized_measure = np.array([])
        vector_weight_aggregated_normalized_measure = np.array([])

        for measure in subcharacteristic["measures"]:
            vector_aggregated_normalized_measure = np.append(vector_aggregated_normalized_measure, measure["value"])
            vector_weight_aggregated_normalized_measure = np.append(
                vector_weight_aggregated_normalized_measure, measure["weight"]
            )
        aggregated_value = calculate_aggregated_weighted_value(
            vector_aggregated_normalized_measure,
            vector_weight_aggregated_normalized_measure,
        )

        result_data["subcharacteristics"].append(
            {
                "key": subcharacteristic_key,
                "value": aggregated_value,
            }
        )

    return result_data


def calculate_characteristics(extracted_characteristics):
    try:
        data = CalculateCharacteristicSchema().load(extracted_characteristics)
    except ValidationError as error:
        raise ValidationError(f"error: Failed to validate input.\nschema_errors: {error.messages}")

    result_data = {"characteristics": []}

    for characteristic in data["characteristics"]:
        characteristic_key: str = characteristic["key"]

        vector_aggregated_normalized_subcharacteristics = np.array([])
        vector_weight_aggregated_normalized_subcharacteristics = np.array([])
        for subcharacteristics in characteristic["subcharacteristics"]:
            vector_aggregated_normalized_subcharacteristics = np.append(
                vector_aggregated_normalized_subcharacteristics,
                subcharacteristics["value"],
            )
            vector_weight_aggregated_normalized_subcharacteristics = np.append(
                vector_weight_aggregated_normalized_subcharacteristics,
                subcharacteristics["weight"],
            )

        aggregated_value = calculate_aggregated_weighted_value(
            vector_aggregated_normalized_subcharacteristics,
            vector_weight_aggregated_normalized_subcharacteristics,
        )

        result_data["characteristics"].append(
            {
                "key": characteristic_key,
                "value": aggregated_value,
            }
        )

    return result_data


def calculate_tsqmi(extracted_tsqmi):
    try:
        data = CalculateTSQMISchema().load(extracted_tsqmi)
    except ValidationError as error:
        raise ValidationError(f"error: Failed to validate input.\nschema_errors: {error.messages}")

    result_data = {"tsqmi": []}

    tsqmi = data["tsqmi"]
    tsqmi_key: str = tsqmi["key"]

    vector_aggregated_normalized_characteristics = np.array([])
    vector_weight_aggregated_normalized_characteristics = np.array([])
    for characteristic in tsqmi["characteristics"]:
        vector_aggregated_normalized_characteristics = np.append(
            vector_aggregated_normalized_characteristics, characteristic["value"]
        )
        vector_weight_aggregated_normalized_characteristics = np.append(
            vector_weight_aggregated_normalized_characteristics,
            characteristic["weight"],
        )

    aggregated_value = calculate_aggregated_weighted_value(
        vector_aggregated_normalized_characteristics,
        vector_weight_aggregated_normalized_characteristics,
    )

    result_data["tsqmi"].append(
        {
            "key": tsqmi_key,
            "value": aggregated_value,
        }
    )

    return result_data



from enum import IntEnum

import numpy as np

from util.check import Checker
from util.exceptions import (
    InvalidGainInterpretationValue,
    ReleasePlannedAndDevelopedOfDifferentSizes,
)


class GainInterpretation(IntEnum):
    NEGATIVE_SLOPE = -1
    POSITIVE_SLOPE = 1


def interpretation_function(
    x: np.array,
    min_threshold: float,
    max_threshold: float,
    gain_interpretation: GainInterpretation,
) -> np.array:
    #  TODO: create docstring in sphinx format

    x = interpolation(x, min_threshold, max_threshold)

    if gain_interpretation == GainInterpretation.POSITIVE_SLOPE:
        return x
    elif gain_interpretation == GainInterpretation.NEGATIVE_SLOPE:
        return 1 - x
    else:
        InvalidGainInterpretationValue(
            f"Gain Interpretation need be 1 or -1, but received {gain_interpretation}"
        )


def interpolation(x: np.array, min: float, max: float) -> np.array:
    """
    Interpolates values of x on x_axis[min, max] and y_axis[0, 1]
    """

    y = np.array([0, 1])
    return np.interp(x, np.array([min, max]), y)


def calculate_measure(interpretated_measure: np.array, number_of_files=None) -> float:
    if number_of_files:
        aggregated_and_normalized_measure = np.divide(
            np.sum(interpretated_measure), number_of_files
        )
        if np.isnan(aggregated_and_normalized_measure) or np.isinf(
            aggregated_and_normalized_measure
        ):
            return 0.0
    else:
        aggregated_and_normalized_measure = interpretated_measure
    return aggregated_and_normalized_measure


def calculate_aggregated_weighted_value(values, weights):
    Checker.check_aggregated_weighted_values(values, weights)

    aggregated_weighted_value = np.linalg.norm(
        np.array(values * weights)
    ) / np.linalg.norm(weights)

    return aggregated_weighted_value


def norm_diff(rp, rd):
    """The norm_diff function performs a vector transformation between the planned release vector and the redeveloped
    release vector. This transformation represents the quantitative perception of the difference between the
    planned quality requirement for a release, and the observed result, after its development.
    """

    if len(rp) != len(rd):
        raise ReleasePlannedAndDevelopedOfDifferentSizes(
            "The size between planned and developed release vectors is not equal.",
        )
    return np.linalg.norm(rp - rd) / np.linalg.norm(rp)


def diff(rp, rd):
    """The diff function interprets a vector transformation between the planned and developed release vectors.
    It generates a vector that expresses whether the result observed in a release is above or below the planned quality
    requirement."""
    if len(rp) != len(rd):
        raise ReleasePlannedAndDevelopedOfDifferentSizes(
            "The size between planned and developed release vectors is not equal.",
        )
    return [max(0, x - y) for x, y in zip(rp, rd)]
